//-< GUESS.CXX >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:      7-Jun-97    K.A. Knizhnik  * / [] \ *
//                          Last update: 17-Oct-97    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Example program: game "Guess an animal"
//-------------------------------------------------------------------*--------*

#include "goods.h"

#define MAX_STRING_SIZE 256

class tree_node : public object {
  public: 
    ref<tree_node> yes;
    ref<tree_node> no;
    char           question[1]; 

    void           guess() const;

    ref<tree_node> ask() const; 

    static tree_node* who(ref<tree_node> const& parent); 

    static tree_node* create(ref<tree_node> const& no,
			     char* question,
			     ref<tree_node> const& yes);


    METACLASS_DECLARATIONS(tree_node, object);

  protected: 
    tree_node(ref<tree_node> const& no, 
	      char* question, 
	      ref<tree_node> const& yes); 
};

tree_node::tree_node(ref<tree_node> const& no, 
		     char* question, 
		     ref<tree_node> const& yes)
: object(self_class, strlen(question)+1)
{
    this->no  = no;
    this->yes = yes;
    strcpy(this->question, question);
}

tree_node* tree_node::create(ref<tree_node> const& no, 
			     char* question, 
			     ref<tree_node> const& yes)
{
    return new (self_class, strlen(question)+1) tree_node(no, question, yes);
}    

void input(char* prompt, char* buf, size_t buf_size)
{
    char* p;
    do { 
	console::output(prompt);
	*buf = '\0';
	console::input(buf, buf_size);
	p = buf + strlen(buf);
    } while (p <= buf+1); 
    
    if (*(p-1) == '\n') {
	*--p = '\0';
    }
}

bool positive_answer()
{
    char buf[16];
    return console::input(buf, sizeof buf) && (*buf == 'y' || *buf == 'Y');
}


tree_node* tree_node::who(ref<tree_node> const& parent) 
{
    char animal[MAX_STRING_SIZE], question[MAX_STRING_SIZE];
    input("What is it ?\n", animal, sizeof animal);
    input("What is difference from other ?\n", question, sizeof question);
    return create(parent, question, create(NULL, animal, NULL));
}

ref<tree_node> tree_node::ask() const
{  
    console::output("May be, %s (y/n) ? ", question);
    if (positive_answer()) { 
	if (yes.is_nil()) { 
	    console::output("It was very simple question for me...\n");
	} else { 
	    ref<tree_node> clarify = yes->ask();
	    if (!clarify.is_nil()) { 
		modify(this)->yes = clarify;
	    }
	}
    } else { 
	if (no.is_nil()) { 
	    if (yes.is_nil()) { 
		return who(this);
	    } else {
		modify(this)->no = who(NULL);
	    } 
	} else { 
	    ref<tree_node> clarify = no->ask();
	    if (!clarify.is_nil()) { 
		modify(this)->no = clarify;
	    }
	}
    }
    return NULL; 
}

void tree_node::guess() const
{
    if (is_abstract_root()) { 
	modify(this)->become(who(NULL));
    } else { 
	ask();
    }
}

field_descriptor& tree_node::describe_components()
{
    return FIELD(yes), FIELD(no), VARYING(question);
}

void on_transaction_abort(metaobject*)
{
    console::output("Let's try again...\n");
}
    
REGISTER(tree_node, object, optimistic_repeatable_read_scheme);


int main() 
{
    task::initialize(task::huge_stack);
    database db;

    char* cfg_name = "guess.cfg";

    if (db.open(cfg_name)) { 
	CLASSOF(storage_root).mop = &optimistic_repeatable_read_scheme; 
	optimistic_repeatable_read_scheme.set_abort_transaction_hook
	    (on_transaction_abort);

	ref<tree_node> root;
	db.get_root(root);

	while(true) { 
	    console::output("Think of an animal.\nReady (y/n) ? ");
	    if (!positive_answer())
	    {
		break;
	    }
	    root->guess();
	}
	console::output("End of the game\n");
	db.close();
	return EXIT_SUCCESS;
    } else { 
	console::output("Failed to open database\n");
	return EXIT_FAILURE;
    }
}

